# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .decimalmeid import DecimalMEID
from .device import Device
from .devicetype import DeviceType
from .imei import IMEI
from .picqeruser import PicqerUser
from .picqerwarehouse import PicqerWarehouse
from .productreference import ProductReference
from .purchaser import Purchaser
from .purchasedproduct import PurchasedProduct
from .supplierreference import SupplierReference
from .udid import UDID


__all__: list[str] = [
    'DecimalMEID',
    'Device',
    'DeviceType',
    'IMEI',
    'PicqerUser',
    'PicqerWarehouse',
    'ProductReference',
    'Purchaser',
    'PurchasedProduct',
    'SupplierReference',
    'UDID',
]


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
from typing import Any

from ckms.core import parse_dsn

AORTA_COMMAND_TOPIC: str = 'molano.commands'

AORTA_COMMAND_PROJECT: str = os.environ['GOOGLE_SERVICE_PROJECT']

AORTA_EVENT_TOPIC: str = 'molano.events'

AORTA_EVENT_PROJECT: str = os.environ['GOOGLE_HOST_PROJECT']

API_BASE_DOMAIN: str = os.environ['API_BASE_DOMAIN']

APP_ENCRYPTION_KEY: str = 'enc'

APP_SIGNING_KEY: str = 'sig'

GOOGLE_DATASTORE_NAMESPACE: str = os.environ['GOOGLE_DATASTORE_NAMESPACE']

GOOGLE_HOST_PROJECT: str = os.environ['GOOGLE_HOST_PROJECT']

GOOGLE_SERVICE_PROJECT: str = os.environ['GOOGLE_SERVICE_PROJECT']

KEYCHAIN: dict[str, Any] = {
    APP_ENCRYPTION_KEY: {
        **parse_dsn(os.environ['APP_ENCRYPTION_KEY']),
        'tags': ['oauth2-client']
    },
    APP_SIGNING_KEY: {
        **parse_dsn(os.environ['APP_SIGNING_KEY']),
        'tags': ['oauth2-client']
    },
}

OAUTH2_SERVER: str = 'https://accounts.webidentity.id'

OAUTH2_SERVICE_CLIENT: str = os.environ['OAUTH_SERVICE_CLIENT']

RESOURCE_SERVERS: dict[str, Any] = {
    'isg': {
        'server': f'https://isg.{API_BASE_DOMAIN}',
        'scope': {"molanoapis.com/internal"}
    },
}


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import functools
from typing import Any
from typing import Callable
from typing import TypeVar

from cbra.ext import ioc
from pydantic import BaseModel
from google.cloud.datastore import Client
from google.cloud.datastore import Entity
from google.cloud.datastore import Key
from google.cloud.datastore import Query


T = TypeVar("T")
M = TypeVar("M")

class GoogleDatastoreRepository:
    __module__: str = 'login.infra.repo'
    client: Client
    exclude: set[str] = set()
    kind: str
    model: type[BaseModel]
    id_attr: str

    @staticmethod
    async def run_in_executor(
        func: Callable[..., Any],
        *args: Any,
        **kwargs: Any
    ) -> Any:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, func, *args, **kwargs)

    def __init__(
        self,
        client: Client = ioc.inject('DatastoreClient')
    ):
        super().__init__()
        self.client = client

    async def allocate(self, base: Key | None = None, kind: str | None = None) -> int:
        if base is None:
            base = self.storage_key(kind=kind)
        key, *_ = await self.run_in_executor(
            functools.partial( # type: ignore
                self.client.allocate_ids, # type: ignore
                incomplete_key=base,
                num_ids=1
            )
        )
        return key.id

    async def delete_key(self, key: Key) -> None:
        await self.run_in_executor(self.client.delete, key) # type: ignore

    def entity_factory(
        self,
        id: int | str | None = None,
        key: Key | None = None,
        kind: str | None = None,
        parent: Key | None = None
    ) -> Entity:
        return Entity(key=key or self.storage_key(kind=kind, id=id, parent=parent))

    def restore(self, entity: Entity | None) -> Any:
        if entity is None:
            return None
        instance = self.model.parse_obj({
            **entity,
            self.id_attr: entity.key.id # type: ignore
        })
        return instance

    def storage_key(
        self,
        id: int | str | None = None,
        kind: str | None = None,
        parent: Key | None = None
    ) -> Key:
        return (
            self.client.key(kind or self.kind, id, parent=parent) # type: ignore
            if id is not None
            else self.client.key(kind or self.kind, parent=parent) # type: ignore
        )

    async def delete(self, entity_id: int):
        assert entity_id is not None # nosec
        await self.run_in_executor(
            self.client.delete, # type: ignore
            self.storage_key(entity_id)
        )

    async def get_entity_by_id(
        self,
        entity_id: int | str,
        kind: str | None = None,
        parent: Key | None = None
    ) -> Entity | None:
        return await self.run_in_executor(
            functools.partial(
                self.client.get, # type: ignore
                key=self.client.key(kind or self.kind, entity_id, parent=parent) # type: ignore
            )
        )

    async def get_model(self, id: Any, kind: str | None = None) -> Any | None:
        entity = await self.get_entity_by_id(id, kind=kind)
        obj = None
        if entity is not None:
            obj = self.restore(entity)
            setattr(obj, self.id_attr, entity.key.id) # type: ignore
        return obj

    async def one(self, query: Query):
        """Run a query that is expected to yield exactly one result."""
        result = None
        for entity in await self.run_in_executor(query.fetch): # type: ignore
            if result is not None: # multiple objects returned
                raise Exception("Multiple entities returned")
            result = entity
        return result

    async def first(self, query: Query):
        """Run a query that is expected to yield at least one result."""
        result = None
        for entity in await self.run_in_executor(query.fetch): # type: ignore
            result = entity
            break
        return result

    async def persist_model(self, obj: T, kind: str | None = None) -> T:
        entity = self.entity_factory(id=getattr(obj, self.id_attr), kind=kind)
        entity.update(obj.dict(exclude=self.exclude | {self.id_attr})) # type: ignore
        await self.put(entity)
        return self.model.parse_obj({ # type: ignore
            **entity,
            self.id_attr: entity.key.id # type: ignore
        })

    async def put(self, entity: Entity) -> Entity:
        await self.run_in_executor(self.client.put, entity) # type: ignore
        assert (entity.key.id or entity.key.name) is not None # type: ignore # nosec
        return entity

    def query(self, kind: str | None = None, ancestor: Key | None = None) -> Query:
        return self.client.query(kind=kind or self.kind, ancestor=ancestor) # type: ignore


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .devicediscovered import DeviceDiscovered
from .phonecheckdeviceconnected import PhonecheckDeviceConnected
from .purchaseorderpurchased import PurchaseOrderPurchased
from .stocklevelchanged import StockLevelChanged


__all__: list[str] = [
    'DeviceDiscovered',
    'PhonecheckDeviceConnected',
    'PurchaseOrderPurchased',
    'StockLevelChanged'
]


# Copyright (C) 2022 Cochise Ruhulessin
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic
from ckms.types import ClaimSet
from cbra.ext.oauth2.types import RFC9068Token
from cbra.ext.oauth2.types import IPrincipal


class ServicePrincipal(IPrincipal, RFC9068Token): # type: ignore
    __module__: str = 'libwebid.lib'
    sub: str

    @pydantic.root_validator(pre=True)
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        client_id = values.get('client_id')
        sub = values.get('sub')
        if sub != client_id:
            raise ValueError("The 'client_id' claim must match the 'sub' claim.")
        return values

    @classmethod
    def fromclaimset(cls, claims: ClaimSet) -> 'ServicePrincipal':
        return cls.parse_obj(claims.dict())

    def get_current_scope(self) -> set[str]:
        """Return the scope that is currrently granted to the principal."""
        return set(self.scope)

    def has_scope(self, scope: str | set[str]) -> bool:
        if isinstance(scope, str):
            scope = {scope}
        return bool(set(self.scope) >= scope)

    def is_authenticated(self) -> bool:
        return True

    def is_service(self) -> bool:
        return True

    def is_subject(self) -> bool:
        return False


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import aorta
import pydantic

from ..purchaser import Purchaser
from ..purchasedproduct import PurchasedProduct
from ..supplierreference import SupplierReference
from ..warehousereference import WarehouseReference


class PurchaseOrderPurchased(aorta.Event):
    id: int = pydantic.Field(
        default=...,
        title="Purchase Order ID",
        description="The Picqer ID of the **Purchase Order**."
    )

    ref: str = pydantic.Field(
        default=...,
        title="Reference",
        description=(
            "The **Purchase Order** number as defined by the procurement "
            "system."
        )
    )

    supplier: SupplierReference = pydantic.Field(
        default=...,
        title="Supplier",
        description=(
            "Describes the **Supplier** with which the **Purchase Order** "
            "was placed."
        )
    )

    supplier_document_id: str | None = pydantic.Field(
        default=None,
        title="Supplier Document ID",
        description=(
            "A document ID retained by the **Supplier** for this specific "
            "**Purchase Order**, such as an order or invoice number."
        )
    )

    warehouse: WarehouseReference = pydantic.Field(
        default=...,
        title="Warehouse",
        description=(
            "The **Warehouse** to which the goods included in the **Purchase "
            "Order** are expected to be shipped."
        )
    )

    purchaser: Purchaser = pydantic.Field(
        default=...,
        title="Purchaser",
        description=(
            "The person that placed the **Purchase Order** with the "
            "**Supplier**."
        )
    )

    products: list[PurchasedProduct] = pydantic.Field(
        default=...,
        title="Products",
        description="The list of products that were ordered."
    )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import aorta
import pydantic

from ..device import Device


class DeviceDiscovered(aorta.Event):
    id: int = pydantic.Field(
        default=...
    )

    device: Device = pydantic.Field(
        default=...
    )


# Copyright (C) 2022 Cochise Ruhulessin
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from cbra.conf import settings
from cbra.ext.oauth2 import RFC9068Principal
from cbra.resource import Resource
from .serviceprincipal import ServicePrincipal


if hasattr(settings, 'OAUTH2_SERVER'):
    TRUSTED_ISSUERS: set[str] = {f"{getattr(settings, 'OAUTH2_SERVER')}"}
else:
    TRUSTED_ISSUERS: set[str] = set() # type: ignore


class ServiceResource(Resource):
    """A :class:`cbra.resource.Resource` implementation that authenticates
    a service with an :rfc:`9068` access token.
    """
    __abstract__: bool = True
    __module__: str = 'libwebid.lib'
    principal: ServicePrincipal
    principal_factory: Any = RFC9068Principal(
        principal_factory=ServicePrincipal.fromclaimset,
        trusted_issuers=TRUSTED_ISSUERS
    )



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime

import pydantic

from .devicetype import DeviceType
from .decimalmeid import DecimalMEID
from .imei import IMEI
from .udid import UDID


class Device(pydantic.BaseModel):
    color: str
    discovered: datetime = pydantic.Field(default_factory=datetime.utcnow)
    battery_serial: str
    imei: list[IMEI] = []
    manufacturer_name: str
    meid: list[DecimalMEID] = []
    model_name: str
    model_number: str
    pesn: list[str] = []
    serial: str
    type: DeviceType
    udid: UDID


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .productreference import ProductReference


class PurchasedProduct(pydantic.BaseModel):
    """Represents a **Product** that is purchased and expected to
    be delivered.
    """
    order_id: int = pydantic.Field(
        default=...,
        title="Order ID",
        description="A reference to the order with which this product was ordered."
    )

    product: ProductReference = pydantic.Field(
        default=...,
        title="Product",
        description="A reference to the **Product** that was ordered."
    )

    amount: int = pydantic.Field(
        default=0,
        title="Amount",
        description="The expected delivery amount."
    )

    delivered: int = pydantic.Field(
        default=0,
        title="Delivered",
        description="The amount that is delivered."
    )